#include <stdlib.h>
#include <string.h>
#include "library.h"
#include "hash.h"
#include "smalloc.h"
#include "vector.h"
#include "object.h"
#include "type.h"
#include "container.h"

static hash_value_t *freenode = NULL;

static hash_value_t *
hash_value_new(object_t *key, void *data)
{
	hash_value_t *value;
	
	if (freenode) {
		value = freenode;
		freenode = freenode->next;
	}
	else {
		value = (hash_value_t *)smalloc(sizeof(hash_value_t));
	}

	value->key = key;
	value->data = data;
	value->next = NULL;

	return value;
}

static void
hash_value_delete(hash_value_t *v)
{
	v->next = freenode;
	freenode = v;
}

void
hash_delete(hash_t *h)
{
	int				 i;
	hash_value_t	*hv;
	hash_value_t	*temp;

	for (i = 0; i < h->size; i++) {
		hv = h->data[i];
		while (hv) {
			if (h->h_delete)
				h->h_delete(hv->data);
			/* I have to do this here, because I don't know
			 * 'h' in hash_value_delete */
			container_remove((container_t *)h, hv->key);
			container_remove((container_t *)h, hv->data);
			temp = hv;
			hv = hv->next;
			hash_value_delete(temp);
		}
	}

	object_unref(OBJ(h->strbuf));
	free(h->data);
	container_delete((container_t *)h);
}

void
hash_init(hash_t *h)
{
	container_init((container_t *)h);
	h->data = (hash_value_t **)smalloc(HASH_SIZE * sizeof(hash_value_t *));
	h->size = HASH_SIZE;
	memset(h->data, 0, HASH_SIZE * sizeof(hash_value_t *));
	h->strbuf = rope_new();
	object_ref(OBJ(h->strbuf));
	h->h_delete = NULL;
	// DPRINTF((stderr, "Returning a new hash table at %p\n", h));
}

/* This is about all I can do again. (See vector) */
char *
hash_stringify_function(hash_t *h)
{
	object_t	*o;
	object_t	*p;
	vector_t	*v;
	rope_t		*r;
	int			 len;
	int			 i;

	r = h->strbuf;

	v = hash_keys(h);
	object_ref(OBJ(v));

	len = vector_length(v);
	
	rope_set(r, "([");

	for (i = 0; i < len; i++) {
		o = vector_data(v, i);
		p = hash_get(h, o);
		rope_add(r, object_stringify(o));
		rope_add(r, ":");
		rope_add(r, object_stringify(p));
		rope_add(r, ",");
	}

	rope_add(r, "])");

	object_unref(OBJ(v));

	return rope_get(r);
}

void
hash_class_init(hash_class_t *c)
{
	container_class_t	*cc;
	object_class_t  	*oc;

	cc = ((container_class_t *)(c));
	container_class_init(cc);

	oc = ((object_class_t *)(c));
	oc->delete = (delete_t)hash_delete;
	oc->string = (stringify_t)hash_stringify_function;
}

int
hash_type(void)
{
    static int   type = 0;

    if (!type) {
        type_t ti = {
            "hash",
            sizeof(hash_t),
            sizeof(hash_class_t),
            (init_t)hash_init,
            (cinit_t)hash_class_init,
            NULL
        };
        type = type_unique(&ti);
    }

    return type;
}

void
hash_set_destructor(hash_t *h, hash_delete_t f)
{
	h->h_delete = f;
}

hash_t *
hash_new(void)
{
	return (hash_t *)type_new(hash_type());
}

/* I need to store hash functions in the objects */
static int
hash_function(hash_t *hashtable, object_t *key)
{
    unsigned int hashvalue = 0;

	hashvalue = object_hash(key);

    hashvalue %= hashtable->size;
    return hashvalue;
}

static hash_value_t *
hash_value_get(hash_t *hash, unsigned int offset, object_t *key)
{
	hash_value_t	*value;
	hash_value_t	*temp = NULL;

	value = hash->data[offset];

	// DPRINTF((stderr, "Value is %p\n", value));
	while (value) {
		if (object_equals(value->key, key) == 0) {
			if (temp) {	/* Move to head of chain */
				// DPRINTF((stderr, "Caching %s\n", key));
				temp->next = value->next;
				value->next = hash->data[offset];
				hash->data[offset] = value;
			}
			return value;
		}

		temp = value;
		value = value->next;
	}

	return NULL;
}

static void
hash_value_add(hash_t *hashtable, int offset, hash_value_t *value)
{
    value->next = hashtable->data[offset];
    hashtable->data[offset] = value;
}

void
hash_add(hash_t *hashtable, object_t *key, object_t *data)
{
    unsigned int	 offset;
    hash_value_t	*hashvalue;

	// DPRINTF((stderr, "Hashing '%s' value '%s'\n", key, (char*)data));
    
    offset = hash_function(hashtable, key);

	hashvalue = hash_value_get(hashtable, offset, key);
	if (hashvalue) {
		// DPRINTF((stderr, "Overwriting %s\n", rope_chars(key)));

		/* They might be the same object (see integer equals()) */
		/* Note that we don't replace the key */
		// container_add((container_t *)hashtable, key);
		// container_remove((container_t *)hashtable, hashvalue->key);
		/* The logic in this is that if one key refs several data
		 * items, we only need one key in memory. If we then replace
		 * one of these, it's more efficient to keep just another
		 * reference on the same key. We must be careful not to frig
		 * any of these keys by reference */

		if (hashtable->h_delete)
			hashtable->h_delete(hashvalue->data);

		container_add((container_t *)hashtable, data);
		container_remove((container_t *)hashtable, hashvalue->data);

		hashvalue->data = data;
	}
	else {
		// DPRINTF((stderr, "Adding %s\n", rope_chars(key)));

		container_add((container_t *)hashtable, key);
		container_add((container_t *)hashtable, data);

		hashvalue = hash_value_new(key, data);
		hash_value_add(hashtable, offset, hashvalue);
	}
}

object_t *
hash_get(hash_t *hashtable, object_t *key)
{
    int				 offset;
	hash_value_t	*val;
    
    offset = hash_function(hashtable, key);

	// DPRINTF((stderr, "Offset is %d\n", offset));
    
    val = hash_value_get(hashtable, offset, key);
	if (!val) return NULL;
	return val->data;
}

bool
hash_remove(hash_t *h, object_t *key)
{
    int				 offset;
	hash_value_t	*val;
    
    offset = hash_function(h, key);

	// DPRINTF((stderr, "Offset is %d\n", offset));

    val = hash_value_get(h, offset, key);

	if (!val) return FALSE;

	/* I know this is at the top of the chain */

	container_remove((container_t *)h, val->key);
	container_remove((container_t *)h, val->data);

	h->data[offset] = val->next;
	hash_value_delete(val);

	return TRUE;
}

vector_t *
hash_keys(hash_t *h)
{
	int				 i;
	hash_value_t	*value;
	vector_t		*v;

	v = vector_new();

	for (i = 0; i < h->size; i++) {
		value = h->data[i];
		while (value) {
			vector_push_back(v, value->key);
			value = value->next;
		}
	}

	return v;
}
