#include "library.h"
#include "list.h"
#include "wregex.h"
#include "regexlist.h"
#include "smalloc.h"

#ifdef DPRINTF
#undef DPRINTF
#define DPRINTF(x) do { } while(0)
#endif

/* This one can be private */
typedef
struct _regexlist_data_t {
	wregex_t	*preg;
	bool		 invert;
} regexlist_data_t;

static regexlist_data_t *
regexlist_data_new(char *regex, bool invert)
{
	regexlist_data_t	*d;
	wregex_t			*w;

	w = wregex_new(regex, REG_ICASE);
	if (!w) return NULL;	/* Nothing allocated, nothing lost */

	d = (regexlist_data_t *)smalloc(sizeof(regexlist_data_t));
	d->preg = w;
	d->invert = !!invert;

	DPRINTF((stderr, "Created regex for '%s' invert %d\n",
					regex, invert));

	return d;
}

static void
regexlist_data_delete(regexlist_data_t *d)
{
	wregex_delete(d->preg);
	sfree(d);
}

regexlist_t *
regexlist_new(void)
{
	regexlist_t	*r;
	r = (regexlist_t *)smalloc(sizeof(regexlist_t));
	r->list = list_new((list_delete_t)regexlist_data_delete);
	r->iter = list_iterator_new(r->list);
	return r;
}

void
regexlist_delete(regexlist_t *r)
{
	list_iterator_delete(r->iter);
	list_delete(r->list);
	sfree(r);
}

bool
regexlist_push_back(regexlist_t *r, char *regex, bool invert)
{
	regexlist_data_t	*d;
	d = regexlist_data_new(regex, invert);
	if (!d) return FALSE;
	list_data_insert(r->iter, regex, (void *)d);
	return TRUE;
}

bool
regexlist_execute(regexlist_t *r, char *str)
{
	regexlist_data_t	*d;
	bool				 result;

	/* Should be true anyhow, but this makes us robust against
	 * changing implementations. */
	/* Going next after an end() returns begin */
	if (list_iterator_end(r->iter)) return TRUE;
	
	while (list_iterator_next(r->iter)) {
		d = (regexlist_data_t *)list_data_read_data(r->iter);
		DPRINTF((stderr, "Got regex for '%s', invert %d\n",
						list_data_read_label(r->iter), d->invert));
		result = wregex_execute(d->preg, str);
		/* If we fail, result == 1, else 0 */
		/* If we invert, invert == 1, else 0 */
		if (result != d->invert) {
			DPRINTF((stderr, "Got match fail between '%s' and '%s'\n",
							str, list_data_read_label(r->iter)));
			return 1;	/* 1 is fail */
		}
	}

	/* See comment above */
	list_iterator_end(r->iter);
	return 0;	/* 0 is success */
}
