#include <stdlib.h>
#include <string.h>
#include "library.h"
#include "smalloc.h"
#include "vector.h"
#include "rope.h"
#include "object.h"
#include "type.h"
#include "container.h"

#define vector_full(v) (v->length == v->size)

void
vector_delete(vector_t *v)
{
	int				 i;
	vector_cell_t	 d;

	for (i = 0; i < vector_length(v); i++) {
		d = vector_data(v, i);
		if (d && v->c_delete)
			v->c_delete(d);
		object_unref(d);
	}

	object_unref(OBJ(v->strbuf));

	sfree(v->cells);
	container_delete((container_t *)v);
}

void
vector_init(vector_t *v)
{
	container_init((container_t *)v);
	v->cells = (vector_cell_t *)smalloc(VECTOR_SIZE * sizeof(vector_cell_t));
	v->size = VECTOR_SIZE;
	v->length = 0;
	v->strbuf = rope_new();
	object_ref(OBJ(v->strbuf));	/* This isn't really IN the container */
	v->c_delete = NULL;
}

/* There's not much I can do about hashing, and pointer equality
 * will do, but I can stringify. */

char *
vector_stringify_function(vector_t *v)
{
	object_t	*o;
	rope_t		*r;
	int			 len;
	int			 i;

	r = v->strbuf;

	len = vector_length(v);
	
	rope_set(r, "({");

	for (i = 0; i < len; i++) {
		o = vector_data(v, i);
		rope_add(r, object_stringify(o));
		rope_add(r, ",");
	}

	rope_add(r, "})");

	return rope_get(r);
}

void
vector_class_init(vector_class_t *c)
{
	container_class_t	*cc;
	object_class_t		*oc;

	cc = ((container_class_t *)(c));
	container_class_init(cc);

	oc = ((object_class_t *)(c));
	oc->delete = (delete_t)vector_delete;
	oc->string = (stringify_t)vector_stringify_function;
}

int
vector_type(void)
{
	static int	 type = 0;

	if (!type) {
		type_t ti = {
			"vector",
			sizeof(vector_t),
			sizeof(vector_class_t),
			(init_t)vector_init,
			(cinit_t)vector_class_init,
			NULL
		};
		type = type_unique(&ti);
	}

	return type;
}

void
vector_set_destructor(vector_t *v, vector_delete_t func)
{
	v->c_delete = func;
}

int
vector_length(vector_t *v)
{
	return v->length;
}

vector_cell_t
vector_data(vector_t *v, int offset)
{
	if (offset > v->length)
		return NULL;
	if (offset < 0)
		return NULL;
#if 0
	DPRINTF((stderr, "Vector data: Size %d, Length %d, Asked %d\n",
					v->size,
					v->length,
					offset));
#endif
	return v->cells[offset];
}

static inline vector_t *
vector_extend(vector_t *v)
{
	if (!vector_full(v)) return v;
	v->size *= 2;
	// DPRINTF((stderr, "Vector size is now %d\n", v->size));
	v->cells = srealloc(v->cells, v->size * sizeof(vector_cell_t));
	return v;
}

#if 0
static inline vector_t *
vector_shrink(vector_t *v)
{
	if (v->length < 9)
		return v;

	if ((v->length * 2) < v->size) {
		v->size /= 2;
		v->cells = (vector_cell_t *)
				srealloc(v->cells, v->size * sizeof(vector_cell_t));
	}
	return v;
}
#endif

void
vector_insert(vector_t *v, vector_cell_t val, int off)
{
	container_add((container_t *)v, val);

	if (vector_full(v))
		vector_extend(v);

	if ((off < 0) || (off > v->length)) {
		DPRINTF((stderr, "Attempt to insert out of range (%d)\n", off));
		return;
	}
	
	if (off < v->length) {
		memmove(
				&(v->cells[(off + 1)]),
				&(v->cells[off]),
				((v->length - off) * sizeof(vector_cell_t))
				);
	}
	v->length++;	/* CAREFUL */
	if (v->length > 1000) {
		DPRINTF((stderr, "Vector length > 1000"));
	}
	v->cells[off] = val;
}

void
vector_push_back(vector_t *v, vector_cell_t val)
{
	vector_insert(v, val, v->length);
}

void
vector_push_front(vector_t *v, vector_cell_t val)
{
	vector_insert(v, val, 0);
}

vector_cell_t
vector_pop(vector_t *v, int off)
{
	vector_cell_t		 val;

	if ((off < 0) || (off > (v->length - 1))) {
		DPRINTF((stderr, "Attempt to pop out of range (%d)\n", off));
		return NULL;
	}

	val = v->cells[off];
	v->length--;
	if (off < v->length) {
		memmove(
				&(v->cells[off]),
				&(v->cells[(off + 1)]),
				((v->length - off) * sizeof(vector_cell_t))
				);
	}

	container_remove((container_t *)v, val);

	return val;
}

vector_cell_t
vector_pop_front(vector_t *v)
{
	return vector_pop(v, 0);
}

vector_cell_t
vector_pop_back(vector_t *v)
{
	return vector_pop(v, v->length - 1);
}

void
vector_remove(vector_t *v, int off)
{
	vector_cell_t	 val;
	val = vector_pop(v, off);
	if (v->c_delete)
		v->c_delete(val);
}

/* Note, compare_t is defined to compare void * */
int
vector_member(vector_t *v, compare_t func, void *val)
{
	int	 len;
	int	 i;

	len = vector_length(v);
	for (i = 0; i < len; i++) {
		if (func(val, vector_data(v, i)) == 0)
			return i;
	}
	return -1;
}

vector_t *
vector_new(void)
{
	return (vector_t *)type_new(vector_type());
}
