#include <stdio.h>
#include <netinet/in.h>
#include <ctype.h>

#include "library.h"
#include "buffer.h"
#include "object.h"
#include "pointer.h"
#include "hash.h"
#include "vector.h"
#include "integer.h"

#include "shandler.h"
#include "network.h"
#include "i3config.h"
#include "i3protocol.h"
#include "i3handler.h"
#include "mudmode.h"

#define I3INFO
#undef I3VERBOSE

static rope_t	*i3router = NULL;	/* Current router */

static hash_t	*i3handlers = NULL;
static vector_t	*i3routers = NULL;
static int		 i3password = CONFIG_PASSWORD;

static hash_t	*i3mudlist = NULL;
static int		 i3mudlist_id = 0;
static hash_t	*i3chanlist = NULL;
static int		 i3chanlist_id = 0;

#define i3handler_print(x) fprintf(stderr, "Object: %s\n", object_stringify(OBJ(x)))

static bool i3handler_packet_send(shandler_t *s, int fd, vector_t *o);
static bool i3handler_packet_header(vector_t *v, int len,
				rope_t **sm, rope_t **su,
				rope_t **tm, rope_t **tu);

/* Remember everything we talk about here is reffed by the vectors
 * and so we don't need to refcount it */

static void
i3handler_startup_reply(shandler_t *s, int fd, vector_t *v)
{
	rope_t		*sm;
	vector_t	*w;
	integer_t	*i;

	if (!i3handler_packet_header(v, 8, &sm, NULL, NULL, NULL))
		return;

	if (object_equals(OBJ(sm), OBJ(i3router))) {
		fprintf(stderr, "WARNING: startup-reply packet received "
						"from %s, not %s. Ignored.\n",
						rope_get(sm), rope_get(i3router));
		return;
	}

	w = (vector_t *)vector_data(v, 6);
	if (!vectorp(w)) {
		fprintf(stderr, "Invalid startup-reply packet - "
						"!vectorp(v[6])\n");
		return;
	}
	i = (integer_t *)vector_data(v, 7);
	if (!integerp(i)) {
		fprintf(stderr, "Invalid startup-reply packet - "
						"!integerp(v[7])\n");
		return;
	}

	object_unref(OBJ(i3routers));
	i3routers = w;
	object_ref(OBJ(i3routers));

	i3password = integer_get(i);

	fprintf(stderr, "===\n");
	fprintf(stderr, "Welcome to the Intermud 3 network\n");
	fprintf(stderr, "Routers: %s\n", object_stringify(OBJ(i3routers)));
	fprintf(stderr, "Password: %d\n", i3password);
	fprintf(stderr, "===\n");

	return;
}

static void
i3handler_mudlist(shandler_t *s, int fd, vector_t *v)
{
	rope_t		*sm;
	hash_t		*h;
	integer_t	*i;
	vector_t	*k;
	vector_t	*d;
	rope_t		*r;
	int			 len;
	int			 off;

	if (!i3handler_packet_header(v, 8, &sm, NULL, NULL, NULL))
		return;

	if (object_equals(OBJ(sm), OBJ(i3router))) {
		fprintf(stderr, "WARNING: mudlist packet received "
						"from %s, not %s. Ignored.\n",
						rope_get(sm), rope_get(i3router));
		return;
	}

	i = (integer_t *)vector_data(v, 6);
	if (!integerp(i)) {
		fprintf(stderr, "Invalid mudlist packet - !integerp(v[6])\n");
		return;
	}
	h = (hash_t *)vector_data(v, 7);
	if (!hashp(h)) {
		fprintf(stderr, "Invalid mudlist packet - !hashp(v[7])\n");
		return;
	}

	i3mudlist_id = integer_get(i);
#ifdef I3VERBOSE
	fprintf(stderr, "New mudlist id is %d\n", i3mudlist_id);
#endif

	k = hash_keys(h);
	len = vector_length(k);
	for (off = 0; off < len; off++) {
		r = (rope_t *)vector_data(k, off);
		if (!ropep(r)) {
			fprintf(stderr, "Invalid MUD name type in mudlist\n");
			i3handler_print(r);
			continue;
		}
		d = (vector_t *)hash_get(h, OBJ(r));
		if (integerp(d)) {
			/* Packet is an integer zero, MUD has gone down */
#ifdef I3VERBOSE
			fprintf(stderr, "MUD %s has zero mapping - down\n",
							rope_get(r));
#endif
			hash_remove(i3mudlist, OBJ(r));
			continue;
		}
		else if (!vectorp(d)) {
			fprintf(stderr, "Invalid MUD data type in mudlist\n");
			i3handler_print(d);
			continue;
		}
		/* XXX I think we assume that this is valid for now, we can
		 * check it when I get around to coding it */
#ifdef I3VERBOSE
		fprintf(stderr, "Added MUD %s to mudlist\n", rope_get(r));
#endif
		hash_add(i3mudlist, OBJ(r), OBJ(d));
	}
}

static void
i3handler_chanlist_reply(shandler_t *s, int fd, vector_t *v)
{
	rope_t		*sm;
	hash_t		*h;
	integer_t	*i;
	vector_t	*k;
	vector_t	*d;
	rope_t		*r;
	int			 len;
	int			 off;

	if (!i3handler_packet_header(v, 8, &sm, NULL, NULL, NULL))
		return;

	if (object_equals(OBJ(sm), OBJ(i3router))) {
		fprintf(stderr, "WARNING: startup-reply packet received "
						"from %s, not %s. Ignored.\n",
						rope_get(sm), rope_get(i3router));
		return;
	}

	i = (integer_t *)vector_data(v, 6);
	if (!integerp(i)) {
		fprintf(stderr, "Invalid chanlist-reply packet - "
						"!integerp(v[6])\n");
		return;
	}
	h = (hash_t *)vector_data(v, 7);
	if (!hashp(h)) {
		fprintf(stderr, "Invalid chanlist-reply packet - "
						"!hashp(v[7])\n");
		return;
	}

	i3chanlist_id = integer_get(i);
#ifdef I3VERBOSE
	fprintf(stderr, "New chanlist id is %d\n", i3chanlist_id);
#endif

	k = hash_keys(h);
	len = vector_length(k);
	for (off = 0; off < len; off++) {
		r = (rope_t *)vector_data(k, off);
		if (!ropep(r)) {
			fprintf(stderr, "Invalid channel name type in chanlist\n");
			i3handler_print(r);
			continue;
		}
		d = (vector_t *)hash_get(h, OBJ(r));
		if (integerp(d)) {
#ifdef I3VERBOSE
			fprintf(stderr, "Channel %d has zero array - removed\n",
							integer_get((integer_t *)d));
#endif
			hash_remove(i3chanlist, OBJ(r));
			continue;
		}
		if (!vectorp(d)) {
			fprintf(stderr, "Invalid channel data type in chanlist\n");
			i3handler_print(d);
			continue;
		}
		/* XXX I think we assume that this is valid for now, we can
		 * check it when I get around to coding it */
#ifdef I3VERBOSE
		fprintf(stderr, "Added channel %s to chanlist\n", rope_get(r));
#endif
		hash_add(i3chanlist, OBJ(r), OBJ(d));
	}
}

static void
i3handler_tell(shandler_t *s, int fd, vector_t *v)
{
	rope_t	*sm;
	rope_t	*su;
	rope_t	*tu;
	rope_t	*d;

	if (!i3handler_packet_header(v, 8, &sm, &su, NULL, &tu))
		return;

	/* Visname ignored */
	d = (rope_t *)vector_data(v, 7);
	if (!ropep(d)) {
		fprintf(stderr, "Invalid tell packet - !ropep(v[6])\n");
		return;
	}

	printf("%s@%s tells %s %s\n",
					rope_get(su), rope_get(sm),
					rope_get(tu), rope_get(d));
}

static void
i3handler_emoteto(shandler_t *s, int fd, vector_t *v)
{
	rope_t	*sm;
	rope_t	*su;
	rope_t	*tu;
	rope_t	*d;

	if (!i3handler_packet_header(v, 8, &sm, &su, NULL, &tu))
		return;

	/* Visname ignored */
	d = (rope_t *)vector_data(v, 7);
	if (!ropep(d)) {
		fprintf(stderr, "Invalid emoteto packet - !ropep(v[6])\n");
		return;
	}

	printf("%s@%s emotes to %s %s\n",
					rope_get(su), rope_get(sm),
					rope_get(tu), rope_get(d));
}

static void
i3handler_who_req(shandler_t *s, int fd, vector_t *v)
{
	vector_t	*w;
	rope_t		*sm;
	rope_t		*su;
	rope_t		*t;

	if (!i3handler_packet_header(v, 6, &sm, &su, NULL, NULL))
		return;

	w = i3protocol_who_reply();
	/* Now frig it by reference, that's what it's for */
	t = (rope_t *)vector_data(w, OFF_TMUD);	/* Target MUD */
	rope_set(t, rope_get(sm));
	t = (rope_t *)vector_data(w, OFF_TUSER);	/* Target user */
	rope_set(t, rope_get(su));

#ifdef I3INFO
	fprintf(stderr, "Processing who request from %s@%s\n",
					rope_get(su), rope_get(sm));
#endif
#ifdef I3VERBOSE
	fprintf(stderr, "Reply is %s\n", object_stringify(OBJ(w)));
#endif

	/* XXX Offset 6 is the data vector - fill it in */
	i3handler_packet_send(s, fd, w);
}

static void
i3handler_who_reply(shandler_t *s, int fd, vector_t *v)
{
	rope_t		*sm;
	rope_t		*tu;

	vector_t	*w;

	vector_t	*x;
	int			 len;
	int			 c;

	rope_t		*n;
	integer_t	*i;
	rope_t		*e;

	if (!i3handler_packet_header(v, 7, &sm, NULL, NULL, &tu))
		return;

	w = (vector_t *)vector_data(v, 6);
	if (!vectorp(v)) {
		fprintf(stderr, "Invalid who-reply packet - !vectorp(v[6])\n");
		return;
	}

	/* XXX find target user */

	/* w is now a vector of who-info vectors */
	len = vector_length(w);
	fprintf(stderr, "Processing who-reply from %s\n", rope_get(sm));
	for (c = 0; c < len; c++) {
		x = (vector_t *)vector_data(w, c);
		if (!vectorp(x)) {
			fprintf(stderr, "Invalid info in who-reply - not vector\n");
			continue;
		}
		if (vector_length(x) != 3) {
			fprintf(stderr, "Invalid info in who-reply - "
							"length != 3\n");
			continue;
		}
		n = (rope_t *)vector_data(x, 0);
		if (!ropep(n)) {
			fprintf(stderr, "Invalid info in who-reply - "
							"!ropep(x[0])\n");
			continue;
		}
		i = (integer_t *)vector_data(x, 1);
		if (!integerp(i)) {
			fprintf(stderr, "Invalid info in who-reply - "
							"!integerp(x[1])\n");
			continue;
		}
		e = (rope_t *)vector_data(x, 2);
		if (!ropep(e)) {
			fprintf(stderr, "Invalid info in who-reply - "
							"!ropep(x[2])\n");
		}
		/* XXX Find user in question */
		fprintf(stderr, "%s (Idle %d) %s\n",
						rope_get(n), integer_get(i), rope_get(e));
	}
}

/* Put in finger stuff here, fit order of documentation */

static void
i3handler_finger_req(shandler_t *s, int fd, vector_t *v)
{
	vector_t	*w;
	rope_t		*l;
	rope_t		*t;
	rope_t		*sm;
	rope_t		*su;

	if (!i3handler_packet_header(v, 7, &sm, &su, NULL, NULL))
		return;

	l = (rope_t *)vector_data(v, 6);
	if (!ropep(l)) {
		fprintf(stderr, "Invalid finger-req packet - !ropep(v[6])\n");
		return;
	}

	/* XXX Now we look up user 'l' and fill in the reply packet */

	w = i3protocol_finger_reply();
	t = (rope_t *)vector_data(w, OFF_TMUD);
	rope_set(t, rope_get(sm));
	t = (rope_t *)vector_data(w, OFF_TUSER);
	rope_set(t, rope_get(su));

#ifdef I3INFO
	fprintf(stderr, "Processing finger request for %s from %s@%s\n",
					rope_get(l), rope_get(su), rope_get(sm));
#endif
#ifdef I3VERBOSE
	fprintf(stderr, "Reply is %s\n", object_stringify(OBJ(w)));
#endif

	i3handler_packet_send(s, fd, w);
}

static void
i3handler_finger_reply(shandler_t *s, int fd, vector_t *v)
{
	rope_t		*sm;
	rope_t		*tu;
	rope_t		*visname;
	rope_t		*title;
	rope_t		*real_name;
	rope_t		*e_mail;
	rope_t		*login_time;
	integer_t	*idle_time;
	rope_t		*ip_name;
	rope_t		*level;
	rope_t		*extra;

	if (!i3handler_packet_header(v, 15, &sm, NULL, NULL, &tu))
		return;

	visname = (rope_t *)vector_data(v, 6);
	if (!ropep(visname)) {
		fprintf(stderr, "Invalid finger-reply packet - !ropep(v[6])\n");
		return;
	}
	title = (rope_t *)vector_data(v, 7);
	if (!ropep(title)) {
		fprintf(stderr, "Invalid finger-reply packet - !ropep(v[7])\n");
		return;
	}
	real_name = (rope_t *)vector_data(v, 8);
	if (!ropep(real_name)) {
		fprintf(stderr, "Invalid finger-reply packet - !ropep(v[8])\n");
		return;
	}
	e_mail = (rope_t *)vector_data(v, 9);
	if (!ropep(e_mail)) {
		fprintf(stderr, "Invalid finger-reply packet - !ropep(v[9])\n");
		return;
	}
	login_time = (rope_t *)vector_data(v, 10);
	if (!ropep(login_time)) {
		fprintf(stderr,"Invalid finger-reply packet - !ropep(v[10])\n");
		return;
	}
	idle_time = (integer_t *)vector_data(v, 11);
	if (!integerp(idle_time)) {
		fprintf(stderr, "Invalid finger-reply packet - "
						"!integerp(v[11])\n");
		return;
	}
	ip_name = (rope_t *)vector_data(v, 12);
	if (!ropep(ip_name)) {
		fprintf(stderr,"Invalid finger-reply packet - !ropep(v[12])\n");
		return;
	}
	level = (rope_t *)vector_data(v, 13);
	if (!ropep(level)) {
		fprintf(stderr,"Invalid finger-reply packet - !ropep(v[13])\n");
		return;
	}
	extra = (rope_t *)vector_data(v, 14);
	if (!ropep(extra)) {
		fprintf(stderr,"Invalid finger-reply packet - !ropep(v[14])\n");
		return;
	}

	/* XXX Find target user */
	fprintf(stderr, "Finger information from %s\n", rope_get(sm));
	fprintf(stderr, "Visname: %s\n", rope_get(visname));
	fprintf(stderr, "Title: %s\n", rope_get(title));
	fprintf(stderr, "Real name: %s\n", rope_get(real_name));
	fprintf(stderr, "E-Mail: %s\n", rope_get(e_mail));
	fprintf(stderr, "Login time: %s\n", rope_get(login_time));
	fprintf(stderr, "Idle time: %d\n", integer_get(idle_time));
	fprintf(stderr, "IP Name: %s\n", rope_get(ip_name));
	fprintf(stderr, "Level: %s\n", rope_get(level));
	fprintf(stderr, "Extra:\n%s\n", rope_get(extra));
}

static void
i3handler_locate_req(shandler_t *s, int fd, vector_t *v)
{
	rope_t	*su;
	rope_t	*sm;
	rope_t	*t;

	if (!i3handler_packet_header(v, 7, &sm, &su, NULL, NULL))
		return;

	t = (rope_t *)vector_data(v, 6);
	if (!ropep(t)) {
		fprintf(stderr, "Invalid locate-req packet = !ropep(v[6])\n");
		return;
	}
	fprintf(stderr, "Locate request for %s from %s@%s\n",
					rope_get(t), rope_get(su), rope_get(sm));
	/* XXX lower case t - probably by user mechanism */
	/* XXX check users */
}

static void
i3handler_locate_reply(shandler_t *s, int fd, vector_t *v)
{
	rope_t		*sm;
	rope_t		*tm;
	rope_t		*tu;
	rope_t		*lm;
	rope_t		*lu;
	integer_t	*i;
	rope_t		*st;

	if (!i3handler_packet_header(v, 10, &sm, NULL, &tm, &tu))
		return;

	lm = (rope_t *)vector_data(v, 6);
	if (!ropep(lm)) {
		fprintf(stderr, "Invalid locate-reply packet - !ropep(v[6])\n");
		return;
	}
	lu = (rope_t *)vector_data(v, 7);
	if (!ropep(lu)) {
		fprintf(stderr, "Invalid locate-reply packet - !ropep(v[7])\n");
		return;
	}
	i = (integer_t *)vector_data(v, 8);
	if (!integerp(i)) {
		fprintf(stderr, "Invalid locate-reply packet - "
						"!integerp(v[8])\n");
		return;
	}
	st = (rope_t *)vector_data(v, 9);
	if (!ropep(st)) {
		fprintf(stderr, "Invalid locate-reply packet - !ropep(v[9])\n");
		return;
	}

	fprintf(stderr, "Locate reply from %s@%s (%d) %s\n",
					rope_get(lu), rope_get(lm),
					integer_get(i), rope_get(st));
	/* XXX Find lu */
}

static void
i3handler_channel_m(shandler_t *s, int fd, vector_t *v)
{
	rope_t	*sm;
	rope_t	*su;
	rope_t	*c;
	rope_t	*n;
	rope_t	*m;

	if (!i3handler_packet_header(v, 9, &sm, &su, NULL, NULL))
		return;
	
	c = (rope_t *)vector_data(v, 6);
	if (!ropep(c)) {
		fprintf(stderr, "Invalid channel-m packet - !ropep(v[6])\n");
		return;
	}
	n = (rope_t *)vector_data(v, 7);
	if (!ropep(n)) {
		fprintf(stderr, "Invalid channel-m packet - !ropep(v[7])\n");
		return;
	}
	m = (rope_t *)vector_data(v, 8);
	if (!ropep(m)) {
		fprintf(stderr, "Invalid channel-m packet - !ropep(v[8])\n");
		return;
	}

	/* Note we ignore visname again */
	fprintf(stderr, "[%s] %s@%s: %s\n",
			rope_get(c), rope_get(su), rope_get(sm), rope_get(m));
}

static void
i3handler_channel_e(shandler_t *s, int fd, vector_t *v)
{
	rope_t	*sm;
	rope_t	*su;
	rope_t	*c;
	rope_t	*n;
	rope_t	*m;

	if (!i3handler_packet_header(v, 9, &sm, &su, NULL, NULL))
		return;
	
	c = (rope_t *)vector_data(v, 6);
	if (!ropep(c)) {
		fprintf(stderr, "Invalid channel-m packet - !ropep(v[6])\n");
		return;
	}
	n = (rope_t *)vector_data(v, 7);
	if (!ropep(n)) {
		fprintf(stderr, "Invalid channel-m packet - !ropep(v[7])\n");
		return;
	}
	m = (rope_t *)vector_data(v, 8);
	if (!ropep(m)) {
		fprintf(stderr, "Invalid channel-m packet - !ropep(v[8])\n");
		return;
	}

	/* Note we ignore visname again */
	fprintf(stderr, "[%s] %s@%s emotes %s\n",
			rope_get(c), rope_get(su), rope_get(sm), rope_get(m));
}

static void
i3handler_channel_t(shandler_t *s, int fd, vector_t *v)
{
	rope_t	*sm;
	rope_t	*su;
	rope_t	*c;
	rope_t	*n;		/* Source visname */

	rope_t	*tmud;	/* Target MUD */
	rope_t	*tuser;	/* Target username */
	rope_t	*moth;	/* Message others */
	rope_t	*mtgt;	/* Message target */
	rope_t	*tvis;	/* Target visname */

	if (!i3handler_packet_header(v, 13, &sm, &su, NULL, NULL))
		return;
	
	c = (rope_t *)vector_data(v, 6);
	if (!ropep(c)) {
		fprintf(stderr, "Invalid channel-m packet - !ropep(v[6])\n");
		return;
	}

	tmud = (rope_t *)vector_data(v, 7);
	if (!ropep(tmud)) {
		fprintf(stderr, "Invalid channel-m packet - !ropep(v[7])\n");
		return;
	}
	tuser = (rope_t *)vector_data(v, 8);
	if (!ropep(tuser)) {
		fprintf(stderr, "Invalid channel-m packet - !ropep(v[8])\n");
		return;
	}
	moth = (rope_t *)vector_data(v, 9);
	if (!ropep(moth)) {
		fprintf(stderr, "Invalid channel-m packet - !ropep(v[9])\n");
		return;
	}
	mtgt = (rope_t *)vector_data(v, 10);
	if (!ropep(mtgt)) {
		fprintf(stderr, "Invalid channel-m packet - !ropep(v[10])\n");
		return;
	}
	n = (rope_t *)vector_data(v, 11);
	if (!ropep(n)) {
		fprintf(stderr, "Invalid channel-m packet - !ropep(v[11])\n");
		return;
	}
	tvis = (rope_t *)vector_data(v, 12);
	if (!ropep(tvis)) {
		fprintf(stderr, "Invalid channel-m packet - !ropep(v[12])\n");
		return;
	}

	/* Note we ignore visname again */
	fprintf(stderr, "[%s] %s@%s emotes to %s@%s %s\n",
			rope_get(c),
			rope_get(su), rope_get(sm),
			rope_get(tuser), rope_get(tmud),
			rope_get(moth));
	fprintf(stderr, "[%s] (For %s@%s: %s\n",
			rope_get(c),
			rope_get(tuser), rope_get(tmud),
			rope_get(mtgt));
}

/* Generics */

static bool
i3handler_packet_header(vector_t *v, int len,
				rope_t **sm, rope_t **su,
				rope_t **tm, rope_t **tu)
{
	object_t	*p;
	char		*pt;
	int			 i;

	if (vector_length(v) != len) {
		fprintf(stderr, "Invalid packet - length != %d\n", len);
		return FALSE;
	}

	p = vector_data(v, 0);	/* Type */
	pt = rope_get((rope_t *)p);

	if (sm) {
		p = vector_data(v, 2);	/* Orig MUD */
		if (!ropep(p)) {
			fprintf(stderr, "Invalid %s packet - p[2] not rope\n", pt);
			return FALSE;
		}
		*sm = (rope_t *)p;
	}

	if (su) {
		p = vector_data(v, 3);	/* Orig user */
		i = object_typenum(p);
		if (!ropep(p)) {
			fprintf(stderr, "Invalid %s packet - p[3] not rope\n", pt);
			return FALSE;
		}
		*su = (rope_t *)p;
	}

	if (tm) {
		p = vector_data(v, 4);	/* Targ MUD */
		i = object_typenum(p);
		if (!ropep(p)) {
			fprintf(stderr, "Invalid %s packet - p[4] not rope\n", pt);
			return FALSE;
		}
		*tm = (rope_t *)p;
	}

	if (tu) {
		p = vector_data(v, 5);	/* Targ user */
		i = object_typenum(p);
		if (!ropep(p)) {
			fprintf(stderr, "Invalid %s packet - p[5] not rope\n", pt);
			return FALSE;
		}
		*tu = (rope_t *)p;
	}

	return TRUE;
}

/* This is pretty minimal now, we do more when we know what we're
 * talking about */
static bool
i3handler_packet_check(object_t *o)
{
	vector_t	*v;
	object_t	*p;

	if (!vectorp(o)) {
		fprintf(stderr, "Invalid packet - not array\n");
		return FALSE;
	}
	v = (vector_t *)o;
	if (vector_length(v) < 6) {
		fprintf(stderr, "Invalid packet - length < 6\n");
		return FALSE;
	}
	p = vector_data(v, 0);	/* Type */
	if (!ropep(p)) {
		fprintf(stderr, "Invalid packet - p[0] not rope\n");
		i3handler_print(p);
		i3handler_print(v);
		return FALSE;
	}
	p = vector_data(v, 1);	/* TTL */
	if (!integerp(p)) {
		fprintf(stderr, "Invalid packet - p[1] not integer\n");
		return FALSE;
	}
	return TRUE;
}

void
i3handler_readcall(shandler_t *s, int fd, buffer_t *b)
{
	int					 blen;
	int					 len;
	byte				*buf;
	object_t			*o;
	pointer_t			*p;
	rope_t				*r;
	i3packethandler_t	 g;

#ifdef I3VERBOSE
	fprintf(stderr, "i3handler_readcall: ...\n");
#endif

	blen = buffer_length(b);

	/* Do we have a network byte order int? */
	while (blen > sizeof(int)) {

		/* Get that int */
		buf = buffer_get(b);
		memcpy(&len, buf, sizeof(int));
		len = ntohl(len);

#ifdef I3VERBOSE
		fprintf(stderr, "Attempting to get packet length %d: ", len);
#endif
		/* Do we have enough for a packet? */
		if (buffer_length(b) < (len + 4)) {
#ifdef I3VERBOSE
			fprintf(stderr, "insufficient data\n");
#endif
			break;
		}

#ifdef I3VERBOSE
		fprintf(stderr, "got packet\n");
#endif

		buf += 4;
		blen -= 4;
		// DPRINTF((stderr, "Buffer content is %s\n", buf));

			/* I hope this stops at the right point */
		o = parse_buffer(buf);
		object_ref(o);


		// printf("Got packet %s\n", object_stringify(o));
		if (i3handler_packet_check(o)) {
			/* XXX Handle packet */
			r = (rope_t *)vector_data((vector_t *)o, 0);
			p = (pointer_t *)hash_get(i3handlers, OBJ(r));
			if (!p) {
				fprintf(stderr, "Packet handler for packet type '%s' "
								"not installed\n", rope_get(r));
				fprintf(stderr, "Buffer is %s\n", buf);
			}
			else {
				g = pointer_get(p);	/* No cast of (void *) */
				g(s, fd, ((vector_t *)(o)));
			}
		}
		else {
			fprintf(stderr, "Packet failed validation - dropped\n");
			fprintf(stderr, "Buffer is %s\n", buf);
		}


		object_unref(o);
		buf += len;
		blen -= len;

		buffer_set(b, buf, blen);
	}

}

static bool
i3handler_packet_send(shandler_t *s, int fd, vector_t *v)
{
	rope_t	*t;
	char	*str;
	int		 len;
	int		 nlen;
	bool	 b;

	t = (rope_t *)vector_data(v, OFF_SUSER);
	if (ropep(t)) {	/* Let's lowercase this */
		str = rope_get(t);
		while (*str) *str = tolower(*str);
		/* We don't need to set this back in as we haven't
		 * modified the length at all */
	}

	str = object_stringify(OBJ(v));
	len = strlen(str);
	nlen = htonl(len);

	fprintf(stderr, "Buffering I3 packet, length %d\n", len);

	b = shandler_write_to(s, fd, (byte *)(&nlen), sizeof(int));
	if (!b) {
		fprintf(stderr, "i3_packet_send: Cannot shandler_write_to "
						"4 bytes\n");
		return FALSE;
	}
	b = shandler_write_to(s, fd, str, len);
	if (!b) {
		fprintf(stderr, "i3_packet_send: Cannot shandler_write_to "
						"%d bytes\n", len);
		return FALSE;
	}
	return TRUE;
}

void
i3handler_add(hash_t *h, char *type, i3packethandler_t g)
{
	rope_t		*r;
	pointer_t	*p;

	r = rope_new_Pc(type);
	p = pointer_new_Pv(g);

	hash_add(h, OBJ(r), OBJ(p));
}

bool
i3handler_boot(void)
{
	if (!i3handlers) {
		i3handlers = hash_new();
		object_ref(OBJ(i3handlers));	/* This is OUR object */

		i3handler_add(i3handlers, "startup-reply",
						i3handler_startup_reply);
		i3handler_add(i3handlers, "mudlist", i3handler_mudlist);
		i3handler_add(i3handlers, "chanlist-reply",
						i3handler_chanlist_reply);
		i3handler_add(i3handlers, "tell", i3handler_tell);
		i3handler_add(i3handlers, "emoteto", i3handler_emoteto);
		i3handler_add(i3handlers, "who-req", i3handler_who_req);
		i3handler_add(i3handlers, "who-reply", i3handler_who_reply);
		i3handler_add(i3handlers, "finger-req", i3handler_finger_req);
		i3handler_add(i3handlers, "finger-reply",
						i3handler_finger_reply);
		i3handler_add(i3handlers, "locate-req", i3handler_locate_req);
		i3handler_add(i3handlers, "locate-reply",
						i3handler_locate_reply);
		i3handler_add(i3handlers, "channel-m", i3handler_channel_m);
		i3handler_add(i3handlers, "channel-e", i3handler_channel_e);
		i3handler_add(i3handlers, "channel-t", i3handler_channel_t);
	}

	if (!i3mudlist) {
		i3mudlist = hash_new();
		object_ref(OBJ(i3mudlist));
	}

	if (!i3routers) {
		i3routers = vector_new();
		object_ref(OBJ(i3routers));
	}

	if (!i3chanlist) {
		i3chanlist = hash_new();
		object_ref(OBJ(i3chanlist));
	}

	return TRUE;
}

bool
i3handler_init(shandler_t *s)
{
	vector_t	*v;
	int			 sock;

	v = i3protocol_startup_req_3();

	i3router = (rope_t *)vector_data(v, OFF_TMUD);
	fprintf(stderr, "Using router name %s for authentication\n",
					rope_get(i3router));

	/* Network init */

	sock = network_connection_new(CONFIG_ROUTER_HOST,
					CONFIG_ROUTER_PORT);
	if (sock == -1) {
		fprintf(stderr, "Could not open connection to I3 router\n");
		return FALSE;
	}

	shandler_add_fd(s, sock, (readcall_t)i3handler_readcall);

	/* Note that we constructed the packets before we opened the
	 * socket, so that we can check for errors before spamming
	 * the network. */

	i3handler_packet_send(s, sock, v);

	return TRUE;
}
