#include <stdio.h>

#include "object.h"
#include "vector.h"
#include "hash.h"
#include "integer.h"
#include "buffer.h"
#include "rope.h"

#include "i3config.h"
#include "i3protocol.h"

/* Careful how you use these, see the comment below */
#define V_SUSER	i3protocol_suser()
#define V_TMUD	i3protocol_tmud()
#define V_TUSER	i3protocol_tuser()
#define V_ROUT	i3protocol_router()
#define V_IDLE	i3protocol_idle()

static vector_t *
i3protocol_header_new(rope_t *type, rope_t *user,
				rope_t *tmud, rope_t *tuser)
{
	static integer_t	*r_ttl = NULL;
	static rope_t		*r_mudname = NULL;
	static integer_t	*r_nouser = NULL;	/* No target user */
	static integer_t	*r_notmud = NULL;	/* No target MUD */

	vector_t			*v;

	if (!r_ttl) {	/* Use as indicator */
		r_ttl = integer_new_i(CONFIG_TTL);
		r_mudname = rope_new_Pc(CONFIG_MUDNAME);
		r_nouser = integer_new_i(0);
		r_notmud = integer_new_i(0);
		object_ref(OBJ(r_ttl));
		object_ref(OBJ(r_mudname));
		object_ref(OBJ(r_nouser));
		object_ref(OBJ(r_notmud));
	}

	/* In this case, we use a new vector each time, as we don't
	 * ref it. It's built up into another packet type. */
	v = vector_new();
	vector_push_back(v, OBJ(type));
	vector_push_back(v, OBJ(r_ttl));
	vector_push_back(v, OBJ(r_mudname));
	if (user)
		vector_push_back(v, OBJ(user));
	else
		vector_push_back(v, OBJ(r_nouser));
	if (tmud) {
		vector_push_back(v, OBJ(tmud));
		if (tuser)
			vector_push_back(v, OBJ(tuser));
		else
			vector_push_back(v, OBJ(r_nouser));
	}
	else {
		vector_push_back(v, OBJ(r_notmud));
		vector_push_back(v, OBJ(r_nouser));
	}

	return v;
}

static hash_t *
i3protocol_services(void)
{
	static hash_t	*h = NULL;

	if (!h) {
		h = hash_new();
		hash_add(h, OBJ(rope_new_Pc("tell")), OBJ(integer_new_i(1)));
		hash_add(h, OBJ(rope_new_Pc("emoteto")), OBJ(integer_new_i(1)));
		hash_add(h, OBJ(rope_new_Pc("who")), OBJ(integer_new_i(1)));
		hash_add(h, OBJ(rope_new_Pc("finger")), OBJ(integer_new_i(1)));
		hash_add(h, OBJ(rope_new_Pc("locate")), OBJ(integer_new_i(1)));
		hash_add(h, OBJ(rope_new_Pc("channel")), OBJ(integer_new_i(1)));
		object_ref(OBJ(h));
	}

	return h;
}

/* Object sharing - not thread safe. These are to be used ONLY when
 * the object's contents will be reset by reference by the handler
 * function. They are NOT to be used when the object content
 * will be shared in more than one place by a vector, or where
 * the data is assumed by the handler function to be correct. */

static rope_t *
i3protocol_suser(void)
{
	static rope_t	*su;

	if (!su) {
		su = rope_new_Pc(CONFIG_USERNAME);
		object_ref(OBJ(su));
	}

	return su;
}

static rope_t *
i3protocol_tmud(void)
{
	static rope_t	*tm;

	if (!tm) {
		tm = rope_new_Pc(CONFIG_MUDNAME);
		object_ref(OBJ(tm));
	}

	return tm;
}

static rope_t *
i3protocol_tuser(void)
{
	static rope_t	*tu;

	if (!tu) {
		tu = rope_new_Pc(CONFIG_USERNAME);
		object_ref(OBJ(tu));
	}

	return tu;
}

static rope_t *
i3protocol_router(void)
{
	static rope_t	*r;

	if (!r) {
		r = rope_new_Pc(CONFIG_ROUTER_NAME);
		object_ref(OBJ(r));
	}

	return r;
}

static object_t *
i3protocol_idle(void)
{
	static integer_t	*i;

	if (!i) {
		i = integer_new_i(666);
		object_ref(OBJ(i));
	}

	return OBJ(i);
}

vector_t *
i3protocol_startup_req_3(void)
{
	static vector_t	*v = NULL;
	rope_t			*t;

	if (!v) {
		t = rope_new_Pc("startup-req-3");
		v = i3protocol_header_new(t, NULL, V_ROUT, NULL);
		/* We ref this, as it's static. t is reffed by v. */
		object_ref(OBJ(v));

			/* password */
		vector_push_back(v, OBJ(integer_new_i(CONFIG_PASSWORD)));
			/* mudlist_id */
		vector_push_back(v, OBJ(integer_new_i(CONFIG_TEMP_MUDLIST_ID)));
			/* chanlist_id */
		vector_push_back(v, OBJ(integer_new_i(CONFIG_TEMP_CHANLIST_ID)));

		vector_push_back(v, OBJ(integer_new_i(CONFIG_PORT_USER)));
		vector_push_back(v, OBJ(integer_new_i(CONFIG_PORT_OOB)));
		vector_push_back(v, OBJ(integer_new_i(CONFIG_PORT_UDP)));
		vector_push_back(v, OBJ(rope_new_Pc(CONFIG_LIB_MAJOR)));
		vector_push_back(v, OBJ(rope_new_Pc(CONFIG_LIB_MINOR)));
		vector_push_back(v, OBJ(rope_new_Pc(CONFIG_DRIVER)));
		vector_push_back(v, OBJ(rope_new_Pc(CONFIG_MUDTYPE)));
		vector_push_back(v, OBJ(rope_new_Pc(CONFIG_STATUS)));
		vector_push_back(v, OBJ(rope_new_Pc(CONFIG_EMAIL)));

		/* Services */
		vector_push_back(v, OBJ(i3protocol_services()));

		vector_push_back(v, OBJ(hash_new()));
	}

	return v;
}

/* These two currently hardcode visname == username by using the
 * same rope. */
vector_t *
i3protocol_tell(void)
{
	static vector_t	*v = NULL;
	rope_t			*t;


	if (!v) {
		t = rope_new_Pc("tell");
		v = i3protocol_header_new(t, V_SUSER, V_TMUD, V_TUSER);

		/* Source visname */
		vector_push_back(v, OBJ(rope_new_Pc(CONFIG_USERNAME)));
		/* Message */
		vector_push_back(v, OBJ(rope_new_Pc("Message")));
		object_ref(OBJ(v));
	}
	
	return v;
}

vector_t *
i3protocol_emoteto(void)
{
	static vector_t	*v = NULL;
	rope_t			*t;

	if (!v) {
		t = rope_new_Pc("emoteto");
		v = i3protocol_header_new(t, V_SUSER, V_TMUD, V_TUSER);

		/* Source visname */
		vector_push_back(v, OBJ(rope_new_Pc(CONFIG_USERNAME)));
		/* Message */
		vector_push_back(v, OBJ(rope_new_Pc("Message")));
		object_ref(OBJ(v));
	}
	
	return v;
}

vector_t *
i3protocol_who_req(void)
{
	static vector_t	*v = NULL;
	rope_t			*t;


	if (!v) {
		t = rope_new_Pc("who-req");
		v = i3protocol_header_new(t, V_SUSER, V_TMUD, NULL);
		object_ref(OBJ(v));
	}
	
	return v;
}

vector_t *
i3protocol_who_reply(void)
{
	static vector_t	*v = NULL;
	rope_t			*t;
	vector_t		*w;
	vector_t		*x;


	if (!v) {
		t = rope_new_Pc("who-reply");
		v = i3protocol_header_new(t, NULL, V_TMUD, V_TUSER);

		/* Build one 'who' entry */
		w = vector_new();
		/* This must be a separate rope */
		/* Username */
		vector_push_back(w, OBJ(rope_new_Pc(CONFIG_USERNAME)));
		vector_push_back(w, V_IDLE);
		vector_push_back(w, OBJ(rope_new_Pc("Idle user")));

		/* Turn this into a vector containing one 'who' entry */
		x = vector_new();
		vector_push_back(x, OBJ(w));

		/* Push onto main vector */
		vector_push_back(v, OBJ(x));

		object_ref(OBJ(v));
	}

	return v;
}

vector_t *
i3protocol_finger_req(void)
{
	static vector_t	*v = NULL;
	rope_t			*t;

	if (!v) {
		t = rope_new_Pc("finger-req");
		v = i3protocol_header_new(t, V_SUSER, V_TMUD, NULL);

		/* User to finger */
		vector_push_back(v, OBJ(rope_new_Pc(CONFIG_USERNAME)));

		object_ref(OBJ(v));
	}

	return v;
}

vector_t *
i3protocol_finger_reply(void)
{
	static vector_t	*v = NULL;
	rope_t			*t;

	if (!v) {
		t = rope_new_Pc("finger-reply");
		v = i3protocol_header_new(t, NULL, V_TMUD, V_TUSER);
		vector_push_back(v, OBJ(rope_new_Pc(CONFIG_USERNAME)));
		vector_push_back(v, OBJ(rope_new_Pc("Title")));
		vector_push_back(v, OBJ(rope_new_Pc("Real name")));
		vector_push_back(v, OBJ(rope_new_Pc("Email")));
		vector_push_back(v, OBJ(rope_new_Pc("Login time")));
		vector_push_back(v, V_IDLE);
		vector_push_back(v, OBJ(rope_new_Pc("Sitename")));
		vector_push_back(v, OBJ(rope_new_Pc("Level")));
		vector_push_back(v, OBJ(rope_new_Pc("Arren is god")));

		object_ref(OBJ(v));
	}

	return v;
}

vector_t *
i3protocol_locate_req(void)
{
	static vector_t	*v = NULL;
	rope_t			*t;

	if (!v) {
		t = rope_new_Pc("locate-req");
		v = i3protocol_header_new(t, V_SUSER, NULL, NULL);
		vector_push_back(v, OBJ(rope_new_Pc(CONFIG_USERNAME)));

		object_ref(OBJ(v));
	}

	return v;

}

vector_t *
i3protocol_locate_reply(void)
{
	static vector_t	*v = NULL;
	rope_t			*t;

	if (!v) {
		t = rope_new_Pc("locate-reply");
		v = i3protocol_header_new(t, NULL, V_TMUD, V_TUSER);
		vector_push_back(v, OBJ(V_TMUD));	/* XXX Beware */
		vector_push_back(v, OBJ(rope_new_Pc(CONFIG_USERNAME)));
		vector_push_back(v, V_IDLE);
		vector_push_back(v, OBJ(rope_new_Pc("NoStatus")));

		object_ref(OBJ(v));
	}

	return v;
}
