%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "library.h"
#include "object.h"
#include "rope.h"
#include "integer.h"
#include "vector.h"
#include "hash.h"
#include "type.h"

#include "mudmode.h"
#include "encode.h"
%}

%option noyywrap
%option always-interactive
%option nounput
%pointer

%x ERROR
%x STRING

dDIGIT		[[:digit:]]
dCHAR		[[:alpha:]]
dWHITE		[[:blank:]]
dESCCHAR	\\\"
dMUDCHAR	[^\"\\]
	/* dSTRING		\"{dMUDCHAR}*\" */
dQUOT		\"
dVALUE		-?{dDIGIT}+
dSPACE		{dWHITE}+
dARRAY		\(\{
dEARRAY		\}\)
dMAPPING	\(\[
dEMAPPING	\]\)
dCOMMA		,
dCOLON		:

%%

{dSPACE}	{ }
{dARRAY}	{ return(MM_ARRAY);		}
{dEARRAY}	{ return(MM_EARRAY);	}
{dMAPPING}	{ return(MM_MAPPING);	}
{dEMAPPING}	{ return(MM_EMAPPING);	}
{dVALUE}	{ return(MM_VALUE);		}
{dCOMMA}	{ return(MM_COMMA);		}
{dCOLON}	{ return(MM_COLON);		}
<*><<EOF>>	{ return(MM_EOF);		}

	/* Strings start with a quote */
{dQUOT}				{ BEGIN(STRING);			}
	/* We can add characters */
<STRING>[^\"\\]+	{ yymore();					}
	/* Trap escaped quotes and other characters */
<STRING>\\.			{ yymore();					}
	/* Strings end with a non-escaped quote */
<STRING>{dQUOT}		{ yytext[(yyleng - 1)] = 0;
					  BEGIN(INITIAL);
					  return MM_STRING;
					}

[^\"]		{ return(MM_ERR);		}

%%

int		 yylval;
bool	 yylerr;

static void
lexerr(const char *fmt, ...)
{
	va_list	 args;
	fprintf(stderr, "Error: ");
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);
	fprintf(stderr, "\n");
	fprintf(stderr, "yytext is '%s', yylval is %d\n", yytext, yylval);
	fflush(stderr);
	yylerr = TRUE;
}

static object_t *
make_rope(char *str)
{
	object_t	*o;
	str = decode_i3(str);
	// DPRINTF((stderr, "Parsed rope at '%s'\n", str));
	o = OBJ(rope_new_Pc(str));
	yylval = yylex();
	return o;
}

static object_t *
make_integer(char *str)
{
	object_t	*o;
	// DPRINTF((stderr, "Parsed integer at '%s'\n", str));
	o = OBJ(integer_new_i(atol(str)));
	yylval = yylex();
	return o;
}

static object_t * parse_value(void);

static object_t *
make_array(char *str)	/* Passed with "({" */
{
	vector_t	*v;
	object_t	*o;

	// DPRINTF((stderr, "Parsed array at '%s'\n", str));

	yylval = yylex();

	v = vector_new();
	
	while (yylval != MM_EARRAY) {
		o = parse_value();
		vector_push_back(v, o);

		switch(yylval) {
			case MM_COMMA:	/* Expected case */
				break;
			case MM_EARRAY:	/* No terminating comma */
				fprintf(stderr, "No terminating comma in array\n");
				goto make_array_end;
			default:
				lexerr("Parse error at %s in array\n", yytext);
				goto make_array_end;
		}

		yylval = yylex();
	}

make_array_end:

	yylval = yylex();
	return OBJ(v);
}

static object_t *
make_mapping(char *str)
{
	hash_t		*h;
	object_t	*o;
	object_t	*p;

	// DPRINTF((stderr, "Parsed mapping at '%s'\n", str));

	yylval = yylex();

	h = hash_new();

	while (yylval != MM_EMAPPING) {
		o = parse_value();

		switch(yylval) {
			case MM_COLON:	/* Expected case */
				break;
			case MM_EMAPPING:
				/* XXX Ditch o */
				goto make_mapping_end;
			default:
				lexerr("Unexpected '%s' in mapping\n", yytext);
				/* XXX Ditch o */
				goto make_mapping_end;
		}

		yylval = yylex();
		p = parse_value();

		hash_add(h, o, p);

		switch(yylval) {
			case MM_COMMA:		/* Expected case */
				break;
			case MM_EMAPPING:	/* No terminating comma */
				fprintf(stderr, "No terminating comma in mapping\n");
				goto make_mapping_end;
			default:
				lexerr("Parse error at %s in mapping\n", yytext);
				goto make_mapping_end;
		}

		/* Lookahead for end or next token ( "])" or "," ) */
		yylval = yylex();
	}

make_mapping_end:

	/* Consume end token */
	yylval = yylex();
	return OBJ(h);
}

static object_t *
parse_value(void)
{
	switch(yylval) {
		case MM_VALUE:
			return make_integer(yytext);
		case MM_STRING:
			return make_rope(yytext);
		case MM_ARRAY:
			return make_array(yytext);
		case MM_MAPPING:
			return make_mapping(yytext);
		default:
			lexerr("Unexpected string %s (%d)\n", yytext, yylval);
			return type_new(object_type());
	}
}

object_t *
parse_buffer(char *buffer)
{
	YY_BUFFER_STATE	 bufstate;
	object_t		*o;

	yylerr = FALSE;

	bufstate = yy_scan_string(buffer);

	yylval = yylex();

	o = parse_value();

	yy_delete_buffer(bufstate);
	return o;
}

#define STR "({9,({4,\"foo\",\"b\\\\\\\"\\\"ar\",-17,}),([6:8,]),})"
// #define STR "({-5,6,7,-19})"

int
mudmode_main(void)
{
	object_t	*o;
	printf("Parsing \"" STR "\"\n");
	o = parse_buffer(STR);
	object_ref(o);
	printf("Stringifies to %s\n", object_stringify(o));
	object_unref(o);
	return 0;
}
