#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>

#include "smalloc.h"
#include "object.h"
#include "buffer.h"

#include "shandler.h"
#include "network.h"

static scallback_t *
scallback_new(int fd, readcall_t read)
{
	scallback_t	*c;
	c = (scallback_t *)smalloc(sizeof(scallback_t));
	c->fd = fd;
	c->read = read;
	c->readbuf = buffer_new();
	c->writebuf = buffer_new();
	c->next = NULL;

	object_ref(OBJ(c->readbuf));
	object_ref(OBJ(c->writebuf));

	return c;
}

static void
scallback_delete(scallback_t *c)
{
	object_unref(OBJ(c->writebuf));
	object_unref(OBJ(c->readbuf));

	sfree(c);
}

/* Now I can go back to classic objects, no inheritance and classes */
shandler_t *
shandler_new(void)
{
	shandler_t	*s;
	s = (shandler_t *)smalloc(sizeof(shandler_t));
	FD_ZERO(&s->rfd);
	FD_ZERO(&s->wfd);
	s->maxfd = 0;
	s->calls = NULL;
	return s;
}

void
shandler_delete(shandler_t *s)
{
	scallback_t	*c;
	scallback_t	*d;

	c = s->calls;
	while (c) {
		d = c->next;
		scallback_delete(c);
		c = d;
	}
	
	sfree(s);
}

static scallback_t *
shandler_get_callback(shandler_t *s, int fd)
{
	scallback_t	*c;
	scallback_t	*d = NULL;

	c = s->calls;
	while (c) {
		if (c->fd == fd) {
			if (d) {
				d->next = c->next;
				c->next = s->calls;
				s->calls = c;
			}
			return c;
		}
		d = c;
		c = c->next;
	}
	return NULL;
}

bool
shandler_add_fd(shandler_t *s, int fd, readcall_t read)
{
	scallback_t	*c;

	if (!read) {
		fprintf(stderr, "Cannot create callback with no read "
						"function - failed\n");
		return FALSE;
	}

	c = shandler_get_callback(s, fd);

	if (c) {
		fprintf(stderr, "Socket %d already exists - not replacing", fd);
		return FALSE;
	}

	c = scallback_new(fd, read);
	c->next = s->calls;
	s->calls = c;

	fprintf(stderr, "Added fd %d to socket handler %p\n", fd, s);
	FD_SET(fd, &s->rfd);
	return TRUE;
}

void
shandler_remove_fd(shandler_t *s, int fd)
{
	scallback_t	*c;

	c = shandler_get_callback(s, fd);

	if (c) {	/* Now it's at the top of the chain */
		s->calls = c->next;
		scallback_delete(c);
	}

	FD_CLR(fd, &s->rfd);
	FD_CLR(fd, &s->rfd);
	FD_CLR(fd, &s->rfd);
}

bool
shandler_write_to(shandler_t *s, int fd, byte *data, int len)
{
	scallback_t	*c;

	c = shandler_get_callback(s, fd);

	if (!c) {
		fprintf(stderr, "shandler_write_to: No such fd %d\n", fd);
		return FALSE;
	}

	buffer_add(c->writebuf, data, len);

	if (buffer_length(c->writebuf) > MAX_WRITEBUF) {
		fprintf(stderr, "shandler_write_to: Write buffer length of %d "
						"exceeded on fd %d\n", MAX_WRITEBUF, fd);
		shandler_remove_fd(s, fd);
		return FALSE;
	}

	// fprintf(stderr, "Setting write flag on fd %d\n", fd);

	FD_SET(fd, &s->wfd);

	if (fd > s->maxfd)
		s->maxfd = fd;

	return TRUE;
}

static void
shandler_read_fd(shandler_t *s, int fd)
{
	scallback_t	*c;
	char		 flags;
	byte		 buf[MAX_READ];
	int			 ret;

	// fprintf(stderr, "shandler_read_fd: ...\n");

	c = shandler_get_callback(s, fd);

	flags = fcntl(fd, F_GETFL, 0);
	fcntl(fd, F_SETFL, flags | O_NONBLOCK);

	ret = read(fd, (void *)buf, MAX_READ);

	fcntl(fd, F_SETFL, flags & ~O_NONBLOCK);

	if (ret <= 0) {
		fprintf(stderr, "Attempt to read from filedescriptor "
						"%d failed\n", fd);
		perror("read");
		/* Remember 0 is EOF */
		if (errno == EAGAIN) {
			fprintf(stderr, "Nonblocking I/O error: No data\n");
		}
		else {
			fprintf(stderr, "Closing socket - EOF\n");
			network_connection_delete(fd);
			fprintf(stderr, "Connection closed\n");
			shandler_remove_fd(s, fd);
			fprintf(stderr, "Connection removed\n");
			return;
		}
	}
	else {
		// fprintf(stderr, "Read %d bytes from fd %d\n", ret, fd);
		buffer_add(c->readbuf, buf, ret);
		if (buffer_length(c->readbuf) > MAX_READBUF) {
			fprintf(stderr, "shandler_read_fd: Read buffer length of %d"
							" exceeded on fd %d\n", MAX_READBUF, fd);
			shandler_remove_fd(s, fd);
			return;
		}
	}

	c->read(s, fd, c->readbuf);
}

static void
shandler_write_fd(shandler_t *s, int fd)
{
	scallback_t	*c;
	byte		*buf;
	int			 len;
	int			 wlen;
	char		 flags;

	fprintf(stderr, "shandler_write_fd: ...\n");

	c = shandler_get_callback(s, fd);

	/* XXX Check c if paranoid */

	flags = fcntl(fd, F_GETFL, 0);
	fcntl(fd, F_SETFL, flags | O_NONBLOCK);

	len = buffer_length(c->writebuf);
	wlen = MIN(len, MAX_WRITE);
	buf = buffer_get(c->writebuf);

	wlen = write(fd, buf, wlen);
	if (len == -1) {
		fprintf(stderr, "Attempt to write to filedescriptor "
						"%d failed\n", fd);
		perror("write");
	}
	else {
		fprintf(stderr, "Wrote %d bytes to fd %d\n", wlen, fd);
		buf += wlen;	/* Move along please */
		len -= wlen;
		buffer_set(c->writebuf, buf, len);
		if (len == 0) {
			// fprintf(stderr, "Buffer for fd %d is empty\n", fd);
			FD_CLR(fd, &s->wfd);
		}
	}

	fcntl(fd, F_SETFL, flags & ~O_NONBLOCK);
}

int
shandler_run(shandler_t *s)
{
	fd_set		 rfd;
	fd_set		 wfd;
	fd_set		 xfd;
	int			 i;

	while (1) {
		memcpy(&rfd, &s->rfd, sizeof(fd_set));
		memcpy(&wfd, &s->wfd, sizeof(fd_set));
		memcpy(&xfd, &s->xfd, sizeof(fd_set));

		// fprintf(stderr, "maxfd for select is %d\n", s->maxfd);
		select((s->maxfd + 1), &rfd, &wfd, &xfd, NULL);
		// fprintf(stderr, "Broke from select\n");

		/* Service the FD */
		for (i = 3; i < (s->maxfd + 1); i++) {

			if (FD_ISSET(i, &xfd)) {
				fprintf(stderr,"Got exception on fd %d. Closing.\n",i);
				network_connection_delete(i);
				shandler_remove_fd(s, i);
				FD_CLR(i, &rfd);
				FD_CLR(i, &wfd);
				FD_CLR(i, &xfd);
			}

			/* Handle writes with buffer */
			if (FD_ISSET(i, &wfd)) {
				shandler_write_fd(s, i);
			}

			if (FD_ISSET(i, &rfd)) {
				shandler_read_fd(s, i);
			}

		}	/* for */
	}		/* while */
}
